#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
#
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 07.07.2023
# Version: 2.0.0
#
# Description:
# This module handles image conversion for the falconz.
#
# Usage:
# The functions in this module can be imported and used in other modules within the falconz to perform image conversion.
#
# ----------------------------------------------------------------------------------------------------------------------

import contextlib
import io
import os
import re
import unicodedata

import SimpleITK
import dicom2nifti
import pydicom
from rich.progress import Progress


def non_nifti_to_nifti(input_path: str, output_directory: str = None) -> None:
    """
    Converts any image format known to ITK to NIFTI

    :param input_path: str, Directory OR filename to convert to nii.gz
    :param output_directory: str, optional output directory to write the image to.
    """

    if not os.path.exists(input_path):
        print(f"Input path {input_path} does not exist.")
        return

    # Processing a directory
    if os.path.isdir(input_path):
        dicom_info = create_dicom_lookup(input_path)
        nifti_dir = dcm2niix(input_path)
        rename_nifti_files(nifti_dir, dicom_info)
        return

    # Processing a file
    if os.path.isfile(input_path):
        # Ignore hidden or already processed files
        _, filename = os.path.split(input_path)
        if filename.startswith('.') or filename.endswith(('.nii.gz', '.nii')):
            return
        else:
            output_image = SimpleITK.ReadImage(input_path)
            output_image_basename = f"{os.path.splitext(filename)[0]}.nii"

    if output_directory is None:
        output_directory = os.path.dirname(input_path)

    output_image_path = os.path.join(output_directory, output_image_basename)
    SimpleITK.WriteImage(output_image, output_image_path)


def standardize_to_nifti(parent_dir: str):
    """
    Converts all images in a parent directory to NIFTI
    """
    # go through the subdirectories
    subjects = os.listdir(parent_dir)
    # get only the directories
    subjects = [subject for subject in subjects if os.path.isdir(os.path.join(parent_dir, subject))]

    with Progress() as progress:
        task = progress.add_task("[white] Processing subjects...", total=len(subjects))
        for subject in subjects:
            subject_path = os.path.join(parent_dir, subject)
            if os.path.isdir(subject_path):
                images = os.listdir(subject_path)
                for image in images:
                    if os.path.isdir(os.path.join(subject_path, image)):
                        image_path = os.path.join(subject_path, image)
                        non_nifti_to_nifti(image_path)
                    elif os.path.isfile(os.path.join(subject_path, image)):
                        image_path = os.path.join(subject_path, image)
                        non_nifti_to_nifti(image_path)
            else:
                continue
            progress.update(task, advance=1, description=f"[white] Processing {subject}...")


def dcm2niix(input_path: str) -> str:
    """
    Converts DICOM images into Nifti images using dcm2niix
    :param input_path: Path to the folder with the dicom files to convert
    :return: str, Path to the folder with the converted nifti files
    """
    output_dir = os.path.dirname(input_path)

    # redirect standard output and standard error to discard output
    with contextlib.redirect_stdout(io.StringIO()), contextlib.redirect_stderr(io.StringIO()):
        dicom2nifti.convert_directory(input_path, output_dir, compression=False, reorient=True)

    return output_dir


def remove_accents(unicode_filename):
    try:
        unicode_filename = str(unicode_filename).replace(" ", "_")
        cleaned_filename = unicodedata.normalize('NFKD', unicode_filename).encode('ASCII', 'ignore').decode('ASCII')
        cleaned_filename = re.sub(r'[^\w\s-]', '', cleaned_filename.strip().lower())
        cleaned_filename = re.sub(r'[-\s]+', '-', cleaned_filename)
        return cleaned_filename
    except:
        return unicode_filename


def is_dicom_file(filename):
    try:
        pydicom.dcmread(filename)
        return True
    except pydicom.errors.InvalidDicomError:
        return False


def create_dicom_lookup(dicom_dir):
    """Create a lookup dictionary from DICOM files.

    Parameters:
    dicom_dir (str): The directory where DICOM files are stored.

    Returns:
    dict: A dictionary where the key is the anticipated filename that dicom2nifti will produce and
          the value is the modality of the DICOM series.
    """

    # a dictionary to store information from the DICOM files
    dicom_info = {}

    # loop over the DICOM files
    for filename in os.listdir(dicom_dir):
        full_path = os.path.join(dicom_dir, filename)
        if is_dicom_file(full_path):
            # read the DICOM file
            ds = pydicom.dcmread(full_path)

            # extract the necessary information
            series_number = ds.SeriesNumber if 'SeriesNumber' in ds else None
            series_description = ds.SeriesDescription if 'SeriesDescription' in ds else None
            sequence_name = ds.SequenceName if 'SequenceName' in ds else None
            protocol_name = ds.ProtocolName if 'ProtocolName' in ds else None
            series_instance_UID = ds.SeriesInstanceUID if 'SeriesInstanceUID' in ds else None
            if ds.Modality == 'PT':
                modality = 'PET'
            else:
                modality = ds.Modality

            # anticipate the filename dicom2nifti will produce and store the modality tag with it
            if series_number is not None:
                base_filename = remove_accents(series_number)
                if series_description is not None:
                    anticipated_filename = f"{base_filename}_{remove_accents(series_description)}.nii"
                elif sequence_name is not None:
                    anticipated_filename = f"{base_filename}_{remove_accents(sequence_name)}.nii"
                elif protocol_name is not None:
                    anticipated_filename = f"{base_filename}_{remove_accents(protocol_name)}.nii"
            else:
                anticipated_filename = f"{remove_accents(series_instance_UID)}.nii"

            dicom_info[anticipated_filename] = modality

    return dicom_info


def rename_nifti_files(nifti_dir, dicom_info):
    """Rename NIfTI files based on a lookup dictionary.

    Parameters:
    nifti_dir (str): The directory where NIfTI files are stored.
    dicom_info (dict): A dictionary where the key is the anticipated filename that dicom2nifti will produce and
                       the value is the modality of the DICOM series.
    """

    # loop over the NIfTI files
    for filename in os.listdir(nifti_dir):
        if filename.endswith('.nii'):
            # get the corresponding DICOM information
            modality = dicom_info.get(filename, '')
            if modality:  # only if the modality is found in the dicom_info dict
                # create the new filename
                new_filename = f"{modality}_{filename}"

                # rename the file
                os.rename(os.path.join(nifti_dir, filename), os.path.join(nifti_dir, new_filename))

                # delete the old name from the dictionary
                del dicom_info[filename]


def split4d(nifti_file: str, out_dir: str) -> None:
    """Split a 4D NIFTI file into 3D NIFTI files using nibabel
    :param nifti_file: 4D NIFTI file to split
    :param out_dir: Directory to save the split NIFTI files
    """
    logging.info(f"Splitting {nifti_file} into 3D nifti files")
    spinner = Halo(text=f"Splitting {nifti_file} into 3D nifti files", spinner='dots')
    spinner.start()
    split_nifti_files = nib.funcs.four_to_three(nib.funcs.squeeze_image(nib.load(nifti_file)))
    i = 0
    for file in split_nifti_files:
        nib.save(file, os.path.join(out_dir, 'vol' + str(i).zfill(4) + '.nii.gz'))
        i += 1
    logging.info(f"Splitting done and split files are saved in {out_dir}")
    spinner.succeed()


def merge3d(nifti_dir: str, wild_card: str, nifti_outfile: str) -> None:
    """
    Merge 3D NIFTI files into a 4D NIFTI file using nibabel
    :param nifti_dir: Directory containing the 3D NIFTI files
    :param wild_card: Wildcard to use to find the 3D NIFTI files
    :param nifti_outfile: User-defined output file name for the 4D NIFTI file
    """
    logging.info(f"Merging 3D nifti files in {nifti_dir} with wildcard {wild_card}")
    files_to_merge = fop.get_files(nifti_dir, wild_card)
    nib.save(nib.funcs.concat_images(files_to_merge, False), nifti_outfile)
    os.chdir(nifti_dir)
    logging.info("Done")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar | Sebastian Gutschmayer
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 04.07.2023
# Version: 0.1.0
#
# Description:
# This module shows predefined display messages for the falconz.
#
# Usage:
# The functions in this module can be imported and used in other modules within the falconz to show predefined display
# messages.
#
# ----------------------------------------------------------------------------------------------------------------------

import pyfiglet
import logging
from falconz import constants
from falconz import file_utilities


def logo():
    """
    Display FALCON logo
    :return:
    """
    print(' ')
    logo_color_code = constants.ANSI_VIOLET
    slogan_color_code = constants.ANSI_VIOLET
    result = logo_color_code + pyfiglet.figlet_format("FALCON 2.0", font="speed").rstrip() + "\033[0m"
    text = slogan_color_code + "A part of the ENHANCE community. Join us at https://enhance.pet to build the future " \
                               "of " \
                               "PET imaging together." + "\033[0m"
    print(result)
    print(text)
    print(' ')


def citation():
    """
    Display manuscript citation
    :return:
    """
    print(
        "Shiyam Sundar, L. K., Lassen, M. L., Muzik, O. (2023, June 8). Fully automated, fast motion correction of "
        "dynamic whole-body and total-body PET/CT Imaging Studies. Journal of Nuclear Medicine. "
        "https://jnm.snmjournals.org/content/early/2023/06/08/jnumed.122.265362 ")
    print(" Copyright 2022, Quantitative Imaging and Medical Physics Team, Medical University of Vienna")


def expectations():
    """
    Display expected modalities for PUMA. This is used to check if the user has provided the correct set of modalities for each tracer set.
    """
    # display the expected modalities
    print(f' Expected dimensions: {constants.EXPECTED_DIMENSIONS} | Allowed modalities: {constants.MODALITIES}')
    logging.info(f' Expected dimensions: {constants.EXPECTED_DIMENSIONS} | Allowed modalities: {constants.MODALITIES}')
    print(
        f"{constants.ANSI_ORANGE} Warning: Only 4D images will be considered in the analysis. {constants.ANSI_RESET}")

    warning_message = "Only 4D images will be considered in the analysis."
    logging.warning(warning_message)


def default_parameters(input_args):
    """
    Display default parameters for FALCON
    :return:
    """
    if input_args.reference_frame_index == -1:
        reference_frame_message = " Reference frame: Last frame"
    else:
        reference_frame_message = f" Reference frame: {input_args.reference_frame_index}"
    if input_args.multi_resolution_iterations.__eq__(constants.MULTI_RESOLUTION_SCHEME):
        multi_resolution_scheme_message = f" Multi-resolution scheme: {constants.MULTI_RESOLUTION_SCHEME}"
    else:
        multi_resolution_scheme_message = f" Multi-resolution scheme: {input_args.multi_resolution_iterations}"

    print(f' Registration type: {input_args.registration} | Shrink level: {constants.SHRINK_LEVEL} | '
          f'{multi_resolution_scheme_message} | {reference_frame_message}')


def derived_parameters(input_args):
    """
    Display derived parameters for FALCON
    :return:
    """
    if input_args.registration.__eq__('rigid'):
        num_jobs, avail_memory, avail_threads = file_utilities.get_number_of_possible_jobs(
            process_memory=c.MINIMUM_RAM_REQUIRED_RIGID,
            process_threads=c.MINIMUM_THREADS_REQUIRED_RIGID)
    elif input_args.registration.__eq__('affine'):
        num_jobs, avail_memory, avail_threads = file_utilities.get_number_of_possible_jobs(
            process_memory=c.MINIMUM_RAM_REQUIRED_AFFINE,
            process_threads=c.MINIMUM_THREADS_REQUIRED_AFFINE)
    elif input_args.registration.__eq__('deformable'):
        num_jobs, avail_memory, avail_threads = file_utilities.get_number_of_possible_jobs(
            process_memory=c.MINIMUM_RAM_REQUIRED_DEFORMABLE,
            process_threads=c.MINIMUM_THREADS_REQUIRED_DEFORMABLE)
    else:
        raise ValueError('Unsupported registration paradigm')

    print(f' Available memory: {avail_memory} GB | Available threads: {avail_threads} | Number of motion correction '
          f'done in parallel: {num_jobs}')
    logging.info(f' Available memory: {avail_memory} GB | Available threads: {avail_threads} | Number of motion '
                 f'correction done in parallel: {num_jobs}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
#
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 07.07.2023
# Version: 2.0.0
#
# Description:
# This module handles input validation for the falconz.
#
# Usage:
# The functions in this module can be imported and used in other modules within the falconz to perform image conversion.
#
# ----------------------------------------------------------------------------------------------------------------------
import os
import logging

class InputValidation:
    def __init__(self, args):
        self.args = args

    def validate(self):
        self._check_directory_exists()
        self._check_reference_frame_index()
        self._check_start_frame()
        self._check_registration_type()
        self._check_multi_resolution_iterations()
        logging.info("Input validation successful.")

    def _check_directory_exists(self):
        if not os.path.exists(self.args.main_folder):
            raise ValueError(f"The specified directory does not exist: {self.args.main_folder}")

    def _check_reference_frame_index(self):
        if not isinstance(self.args.reference_frame_index, int) or self.args.reference_frame_index < -1:
            raise ValueError(
                f"Reference frame index must be a non-negative integer or -1: {self.args.reference_frame_index}")

    def _check_start_frame(self):
        if not isinstance(self.args.start_frame, int) or self.args.start_frame < 0:
            raise ValueError(f"Start frame must be a non-negative integer: {self.args.start_frame}")

    def _check_registration_type(self):
        allowed_registration_paradigms = ["rigid", "affine", "deformable"]
        if self.args.registration not in allowed_registration_paradigms:
            raise ValueError(f"Invalid registration type. Allowed values are: {allowed_registration_paradigms}")

    def _check_multi_resolution_iterations(self):
        if not isinstance(self.args.multi_resolution_iterations, str) or not all(
                i.isdigit() for i in self.args.multi_resolution_iterations.split('x')):
            raise ValueError(
                f"Multi resolution iterations must be a 'x' separated numeric string: {self.args.multi_resolution_iterations}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 04.07.2023
# Version: 0.1.0
#
# Description:
# The main module of the falconz. It contains the main function that is executed when the falconz is run.
#
# Usage:
# The variables in this module can be imported and used in other modules within the falconz.
#
# ----------------------------------------------------------------------------------------------------------------------


import argparse
import logging
import os
import sys
import time
from datetime import datetime
import colorama
import emoji

from falconz import display
from falconz import constants
from falconz import file_utilities
from falconz import download
from falconz import resources
from falconz import image_conversion

logging.basicConfig(format='%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s', level=logging.INFO,
                    filename=datetime.now().strftime('falconz-v.1.0.0.%H-%M-%d-%m-%Y.log'),
                    filemode='w')


def main():
    colorama.init()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-m",
        "--main_folder",
        type=str,
        help="path containing the images to motion correct",
        required=True,
    )
    parser.add_argument(
        "-rf",
        "--reference_frame_index",
        type=int,
        default=-1,
        help="index of the reference frame [index starts from 0]",
    )
    parser.add_argument(
        "-sf",
        "--start_frame",
        type=int,
        default=99,
        help="frame from which the motion correction will be performed"
    )
    parser.add_argument(
        "-r",
        "--registration",
        type=str,
        choices=constants.ALLOWED_REGISTRATION_PARADIGMS,
        required=True,
        help="Type of registration: rigid  | affine  | deformable"
    )
    parser.add_argument(
        "-i",
        "--multi_resolution_iterations",
        type=str,
        default=constants.MULTI_RESOLUTION_SCHEME,
        help="Number of iterations for each resolution level"
    )
    args = parser.parse_args()
    validator = InputValidation(args)
    validator.validate()

    display.logo()
    display.citation()

    logging.info('----------------------------------------------------------------------------------------------------')
    logging.info('                                     STARTING FALCON-Z V.2.0.0                                      ')
    logging.info('----------------------------------------------------------------------------------------------------')

    # ----------------------------------
    # INPUT VALIDATION AND PREPARATION
    # ----------------------------------

    print(' ')
    print(f'{constants.ANSI_VIOLET} {emoji.emojize(":memo:")} NOTE:{constants.ANSI_RESET}')
    print(' ')
    display.expectations()
    display.default_parameters(args)
    display.derived_parameters(args)

    # ----------------------------------
    # DOWNLOADING THE BINARIES
    # ----------------------------------

    print('')
    print(f'{constants.ANSI_VIOLET} {emoji.emojize(":globe_with_meridians:")} BINARIES DOWNLOAD:{constants.ANSI_RESET}')

    print('')
    binary_path = constants.BINARY_PATH
    file_utilities.create_directory(binary_path)
    system_os, system_arch = file_utilities.get_system()
    print(f'{constants.ANSI_ORANGE} Detected system: {system_os} | Detected architecture: {system_arch}'
          f'{constants.ANSI_RESET}')
    download.download(item_name=f'greedy-{system_os}-{system_arch}', item_path=binary_path,
                      item_dict=resources.GREEDY_BINARIES)
    file_utilities.set_permissions(constants.GREEDY_PATH, system_os)

    # ----------------------------------
    # INPUT STANDARDIZATION
    # ----------------------------------

    print('')
    print(f'{constants.ANSI_VIOLET} {emoji.emojize(":magnifying_glass_tilted_left:")} STANDARDIZING INPUT DATA TO '
          f'NIFTI:{constants.ANSI_RESET}')
    print('')
    logging.info(' ')
    logging.info(' STANDARDIZING INPUT DATA TO NIFTI:')
    logging.info(' ')
    image_conversion.standardize_to_nifti(image_directory)
    print(f"{constants.ANSI_GREEN} Standardization complete.{constants.ANSI_RESET}")
    logging.info(" Standardization complete.")

    # ----------------------------------


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 04.07.2023
# Version: 0.1.0
#
# Description:
# This module contains the constants that are used in the falconz.
#
# Usage:
# The variables in this module can be imported and used in other modules within the falconz.
#
# ----------------------------------------------------------------------------------------------------------

import os
from falconz import file_utilities
from datetime import datetime

project_root = file_utilities.get_virtual_env_root()
BINARY_PATH = os.path.join(project_root, 'bin')

# SET PATHS TO BINARIES

if file_utilities.get_system()[0] == 'windows':
    GREEDY_PATH = os.path.join(BINARY_PATH, f'greedy-{file_utilities.get_system()[0]}-{file_utilities.get_system()[1]}',
                               'greedy.exe')
elif file_utilities.get_system()[0] in ['linux', 'mac']:
    GREEDY_PATH = os.path.join(BINARY_PATH, f'greedy-{file_utilities.get_system()[0]}-{file_utilities.get_system()[1]}',
                               'greedy')
else:
    raise ValueError('Unsupported OS')


# COLOR CODES
ANSI_ORANGE = '\033[38;5;208m'
ANSI_GREEN = '\033[38;5;40m'
ANSI_VIOLET = '\033[38;5;141m'
ANSI_RESET = '\033[0m'


# ALLOWED MODALITIES

MODALITIES = ['PET', 'CT', 'MR']

# SYSTEM REQUIREMENTS

MINIMUM_RAM_REQUIRED_RIGID = 4  # in GB
MINIMUM_RAM_REQUIRED_AFFINE = 8  # in GB
MINIMUM_RAM_REQUIRED_DEFORMABLE = 16  # in GB
MINIMUM_THREADS_REQUIRED_RIGID = 2  # in number of threads
MINIMUM_THREADS_REQUIRED_AFFINE = 4  # in number of threads
MINIMUM_THREADS_REQUIRED_DEFORMABLE = 8  # in number of threads

# Shrink levels supported:
SHRINK_LEVEL = [2, 4, 8]
NCC_THRESHOLD = 0.6  # Normalized cross correlation threshold
NCC_RADIUS = '4x4x4'  # Normalized cross correlation radius

# FILE NAMES

MOCO_PREFIX = 'moco_'
ALIGNED_PREFIX = 'aligned_'


# FOLDER NAMES

FALCON_WORKING_FOLDER = 'FALCONZ-V02'+'-' + datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
TRANSFORMS_FOLDER = 'transforms'
MOCO_FOLDER = 'Motion-corrected-images'

# HYPERPARAMETERS

MULTI_RESOLUTION_SCHEME = '100x25x10'
EXPECTED_DIMENSIONS = 4
ALLOWED_REGISTRATION_PARADIGMS = ["rigid", "affine", "deformable"]
IMAGE_INTERPOLATION = 'Linear'
MASK_INTERPOLATION = 'Nearest Neighbor'


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar
#
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 07.07.2023
# Version: 2.0.0
#
# Description:
# This module handles image processing for the falconz.
#
# Usage:
# The functions in this module can be imported and used in other modules within the falconz to perform image conversion.
#
# ----------------------------------------------------------------------------------------------------------------------

import os
import pathlib
import re
import subprocess
import logging
import multiprocessing
from tqdm import tqdm
import SimpleITK
import nibabel
import pandas as pd
from nilearn.input_data import NiftiMasker


class GreedyException(Exception):
    pass


class GreedyCommand:
    def __init__(self, cmd_name, fixed_img, moving_img, cost_function, multi_resolution_iterations, dof, out_file):
        self.cmd_name = cmd_name
        self.fixed_img = fixed_img
        self.moving_img = moving_img
        self.cost_function = cost_function
        self.multi_resolution_iterations = multi_resolution_iterations
        self.dof = dof
        self.out_file = out_file

    def _generate_command(self):
        return f"greedy -d 3 -a -i {re.escape(self.fixed_img)} {re.escape(self.moving_img)} -ia-image-centers -dof {self.dof} -o {re.escape(self.out_file)} -n {self.multi_resolution_iterations} -m {self.cost_function}"

    def run(self):
        cmd = self._generate_command()
        result = subprocess.run(cmd, shell=True, capture_output=True)

        if result.returncode != 0:
            raise GreedyException(f"Command '{cmd}' failed with error: {result.stderr}")

        logging.info(
            f"{self.cmd_name} alignment: {pathlib.Path(self.moving_img).name} -> {pathlib.Path(self.fixed_img).name} | Aligned image: moco-{pathlib.Path(self.moving_img).name} | Cost function: {self.cost_function} | Initial alignment: Image centers | Transform file: {pathlib.Path(self.out_file).name}")

        return self.out_file


class Registration:
    def __init__(self, fixed_img, moving_img, cost_function, multi_resolution_iterations):
        self.fixed_img = fixed_img
        self.moving_img = moving_img
        self.cost_function = cost_function
        self.multi_resolution_iterations = multi_resolution_iterations
        self.out_dir = pathlib.Path(moving_img).parent
        self.moving_img_filename = pathlib.Path(moving_img).name

    def register(self):
        raise NotImplementedError


class RigidRegistration(Registration):
    def register(self):
        rigid_transform_file = os.path.join(self.out_dir, f"{self.moving_img_filename}_rigid.mat")
        cmd = GreedyCommand("Rigid", self.fixed_img, self.moving_img, self.cost_function,
                            self.multi_resolution_iterations, 6, rigid_transform_file)
        return cmd.run()


class AffineRegistration(Registration):
    def register(self):
        affine_transform_file = os.path.join(self.out_dir, f"{self.moving_img_filename}_affine.mat")
        cmd = GreedyCommand("Affine", self.fixed_img, self.moving_img, self.cost_function,
                            self.multi_resolution_iterations, 12, affine_transform_file)
        return cmd.run()


class DeformableRegistration(Registration):
    def register(self):
        warp_file = os.path.join(self.out_dir, f"{self.moving_img_filename}_warp.nii.gz")
        inverse_warp_file = os.path.join(self.out_dir, f"{self.moving_img_filename}_inverse_warp.nii.gz")
        affine_transform_file = AffineRegistration(self.fixed_img, self.moving_img, self.cost_function,
                                                   self.multi_resolution_iterations).register()
        cmd = GreedyCommand("Deformable", self.fixed_img, self.moving_img, self.cost_function,
                            self.multi_resolution_iterations, 6, warp_file)
        return cmd.run()


class ImageResampler:
    def __init__(self, fixed_img, moving_img, resampled_moving_img, registration_type, segmentation="",
                 resampled_seg=""):
        self.fixed_img = fixed_img
        self.moving_img = moving_img
        self.resampled_moving_img = resampled_moving_img
        self.registration_type = registration_type
        self.segmentation = segmentation
        self.resampled_seg = resampled_seg
        self.out_dir = pathlib.Path(moving_img).parent
        self.moving_img_file = pathlib.Path(moving_img).name

    def resample(self):
        if self.registration_type == 'rigid':
            transform_file = os.path.join(self.out_dir, f"{self.moving_img_file}_rigid.mat")
        elif self.registration_type == 'affine':
            transform_file = os.path.join(self.out_dir, f"{self.moving_img_file}_affine.mat")
        elif self.registration_type == 'deformable':
            transform_file = os.path.join(self.out_dir, f"{self.moving_img_file}_warp.nii.gz")
        else:
            raise GreedyException("Registration type not supported!")

        if self.segmentation and self.resampled_seg:
            cmd_to_run = f"greedy -d 3 -rf {re.escape(self.fixed_img)} -ri LINEAR -rm {re.escape(self.moving_img)} " \
                         f"{re.escape(self.resampled_moving_img)} -ri LABEL 0.2vox -rm {re.escape(self.segmentation)} " \
                         f"{re.escape(self.resampled_seg)} -r {re.escape(transform_file)}"
        else:
            cmd_to_run = f"greedy -d 3 -rf {re.escape(self.fixed_img)} -ri LINEAR -rm {re.escape(self.moving_img)} " \
                         f"{re.escape(self.resampled_moving_img)} -r {re.escape(transform_file)}"

        subprocess.run(cmd_to_run, shell=True, capture_output=True)


class ImageAligner:
    def __init__(self, n_jobs):
        self.n_jobs = n_jobs

    def align(self, fixed_img, moving_imgs, registration_type, multi_resolution_iterations, moco_dir):
        logging.info("Aligning images...")

        with multiprocessing.Pool(self.n_jobs) as p:
            max_ = len(moving_imgs)
            with tqdm(total=max_, bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt}', ncols=70, colour='violet') as pbar:
                for i, _ in tqdm(enumerate(p.imap_unordered(self.align_single, [
                    (fixed_img, img, registration_type, multi_resolution_iterations, moco_dir) for img in
                    moving_imgs]))):
                    pbar.update()
                    if pbar.n == max_:
                        pbar.colour = 'green'

    def align_single(self, align_param):
        fixed_img, moving_img, registration_type, multi_resolution_iterations, moco_dir = align_param

        if registration_type == 'rigid':
            registration = RigidRegistration(fixed_img, moving_img, 'NMI', multi_resolution_iterations)
        elif registration_type == 'affine':
            registration = AffineRegistration(fixed_img, moving_img, 'NMI', multi_resolution_iterations)
        elif registration_type == 'deformable':
            registration = DeformableRegistration(fixed_img, moving_img, 'NCC 2x2x2', multi_resolution_iterations)
        else:
            raise GreedyException("Registration type not supported!")

        registration.register()

        resampler = ImageResampler(fixed_img, moving_img,
                                   os.path.join(moco_dir, 'moco-' + pathlib.Path(moving_img).name), registration_type)
        resampler.resample()


def get_dimensions(nifti_file: str) -> int:
    """
    Get the dimensions of a NIFTI image file
    :param nifti_file: NIFTI file to check
    """
    nifti_img = SimpleITK.ReadImage(nifti_file)
    img_dim = nifti_img.GetDimension()
    return img_dim


def get_pixel_id_type(nifti_file: str) -> str:
    """
    Get the pixel id type of a NIFTI image file
    :param nifti_file: NIFTI file to check
    """
    nifti_img = SimpleITK.ReadImage(nifti_file)
    pixel_id_type = nifti_img.GetPixelIDTypeAsString()
    return pixel_id_type


def get_intensity_statistics(nifti_file: str, multi_label_file: str) -> object:
    """
    Get the intensity statistics of a NIFTI image file
    :param nifti_file: NIFTI file to check
    :param multi_label_file: Multilabel file that is used to calculate the intensity statistics from nifti_file
    :return: stats_df, a dataframe with the intensity statistics
    """
    nifti_img = SimpleITK.ReadImage(nifti_file)
    multi_label_img = SimpleITK.ReadImage(multi_label_file)
    intensity_statistics = SimpleITK.LabelIntensityStatisticsImageFilter()
    intensity_statistics.Execute(multi_label_img, nifti_img)
    stats_list = [(intensity_statistics.GetMean(i), intensity_statistics.GetStandardDeviation(i),
                   intensity_statistics.GetMedian(i), intensity_statistics.GetMaximum(i),
                   intensity_statistics.GetMinimum(i)) for i in intensity_statistics.GetLabels()]
    columns = ['Mean', 'Standard Deviation', 'Median', 'Maximum', 'Minimum']
    stats_df = pd.DataFrame(data=stats_list, index=intensity_statistics.GetLabels(), columns=columns)
    return stats_df


def get_body_mask(nifti_file: str, mask_file: str) -> str:
    """
    Get time activity curves from a 4d nifti file
    :param nifti_file: 4d nifti file to get the time activity curves from
    :param mask_file: Name of the mask file that is derived from the 4d nifti file.
    :return: path of the mask file
    """
    nifti_masker = NiftiMasker(mask_strategy='epi', memory="nilearn_cache", memory_level=2, smoothing_fwhm=8)
    nifti_masker.fit(nifti_file)
    nibabel.save(nifti_masker.mask_img_, mask_file)
    return mask_file


def mask_img(nifti_file: str, mask_file: str, masked_file: str) -> str:
    """
    Mask a NIFTI image file with a mask file
    :param nifti_file: NIFTI file to mask
    :param mask_file: Mask file to mask the nifti_file with
    :param masked_file: Name of the masked file
    :return: path of the masked nifti file
    """
    img = SimpleITK.ReadImage(nifti_file)
    mask = SimpleITK.ReadImage(mask_file, SimpleITK.sitkFloat32)
    masked_img = SimpleITK.Compose(
        [SimpleITK.Multiply(SimpleITK.VectorIndexSelectionCast(img, i), mask) for i in
         range(img.GetNumberOfComponentsPerPixel())])
    SimpleITK.WriteImage(SimpleITK.Cast(masked_img, SimpleITK.sitkVectorFloat32), masked_file)
    return masked_file



from setuptools import setup, find_packages

setup(
    name='falconz',
    version='2.0.3',
    author='Lalith Kumar Shiyam Sundar',
    author_email='Lalith.shiyamsundar@meduniwien.ac.at',
    description='FalconZ: A streamlined Python package for PET motion correction.',
    python_requires='>=3.9',
    long_description='FalconZ is a robust and comprehensive Python package that offers a simplified approach to PET ('
                     'Positron Emission Tomography) motion correction. The software is equipped to handle both head '
                     'and total-body scans, ensuring high-accuracy results in diverse settings. Built around the '
                     'potent greedy registration toolkit, FalconZ serves as an efficient engine for registration '
                     'tasks, providing precise alignment in PET imaging. The package has been designed with '
                     'user-friendliness in mind, streamlining the implementation process of motion correction. '
                     'Whether you are a researcher, healthcare professional, or an individual working with PET scans, '
                     'FalconZ provides a seamless solution to your needs. It integrates effortlessly into your '
                     'existing Python environment, and its straightforward usability makes it an indispensable tool '
                     'for accurate PET scan analysis.',
    url='https://github.com/QIMP-Team/FALCON',
    license='GPLv3',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Healthcare Industry',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
    ],

    keywords='PET motion correction, diffeomorphic imaging, image processing',
    packages=find_packages(),
    install_requires=[
        'nibabel~=3.2.2',
        'halo~=0.0.31',
        'SimpleITK~=2.2.1',
        'pydicom~=2.2.2',
        'argparse~=1.4.0',
        'numpy~=1.22.3',
        'mpire~=2.3.3',
        'openpyxl~=3.0.9',
        'matplotlib',
        'pyfiglet~=0.8.post1',
        'natsort~=8.1.0',
        'pillow>=9.2.0',
        'colorama~=0.4.6',
        'rich',
        'pandas',
        'dicom2nifti~=2.4.8',
        'requests',
        'emoji',
    ],
    entry_points={
        'console_scripts': [
            'falconz=falconz.falconz:main',
        ],
    },
)
